
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#define _GNU_SOURCE
#include <fcntl.h>
#include <sys/file.h>
#include <sys/types.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <getopt.h>
#include <sys/socket.h>
#include <sys/un.h>

/* #include "ngc.h" */
#include "../../src/initng.h"
#include "../../src/initng_active_status.h"
#include "initng_ngc2.h"

#define TRUE 1
#define FALSE 0
int debug = 0;

/* Global comunication socket */
static int sock = -1;

/* functions */
void print_usage(int v);
int open_socket(void);
int close_socket(void);
void send_command(const char *command);
char *command_fetch_string(char *got_from_read);
int command_fetch_int(const char *command);
int print_status(char *servicename, e_a_status status, struct tm *ts);
int main(int argc, char *argv[]);

int ngc_get_pid(char *service);
int ngc_take_up(char *service);
int ngc_take_down(char *service);
int ngc_restart_service(char *service);
int ngc_father_to(char *service);
int ngc_reload(char *service);
int ngc_restart_from(char *service);
int ngc_zap(char *service);
int ngc_history(char *opt);
int ngc_status(char *service);
int ngc_restart(char *opt);
int ngc_quit(char *opt);
int ngc_print(char *service);
int ngc_print_active(char *service);
int ngc_percent(char *opt);
int ngc_toggle_verbose(char *opt);
int ngc_s_l(char *opt);
int ngc_c_a_fd(char *opt);
int ngc_hlp(char *opt);
int ngc_m_list(char *opt);
int ngc_m_load(char *opt);
int ngc_m_unload(char *opt);
int ngc_debug_add(char *string);
int ngc_debug_del(char *string);
int ngc_a(char *service);
int ngc_b(char *service);
int ngc_f_r(char *service);


void bad_syntax(void);
char *Argv;

typedef struct
{
    const char *l;              /* long name */
    const char s;               /* short name */
    int a;                      /* uses an option */
    int b;                      /* requires an option */
    int v;                      /* include in help */
    const char *h;              /* help */
    int (*exec) (char *opt);
} opt;



static opt o[] = {
    {"depon", 'a', 1, 1, 0, "What does this service depends on", &ngc_a},
    {"print-active", 'A', 1, 0, 1, "Print active data", &ngc_print_active},
    {"depoff", 'b', 1, 1, 0, "What services does depend on me", &ngc_b},
    {"down", 'd', 1, 1, 1, "stop service", &ngc_take_down},
    {"father", 'f', 1, 1, 1, "get father of", &ngc_father_to},
    {"get", 'g', 1, 1, 1, "output just the PID.", &ngc_get_pid},
    {"help", 'h', 0, 0, 1, "Print Full Help", &ngc_hlp},
    {"debug-add", 'i', 1, 1, 0, "Add an word, that will be verbose.",
     &ngc_debug_add},
    {"startfrom", 'j', 1, 1, 0, "Start over from another service",
     &ngc_restart_from},
    {"debug-del", 'k', 1, 1, 0, "Delete an word, that will be verbose.",
     &ngc_debug_del},
    {"log", 'l', 1, 0, 1, "Output History Log", &ngc_history},
    {"list-modules", 'm', 0, 0, 0, "List all loaded modules", &ngc_m_list},
    {"percent", 'n', 0, 0, 0, "Return current up status in %", &ngc_percent},
    {"load-module", 'o', 1, 1, 0, "Load a module into initng", &ngc_m_load},
    {"print", 'p', 1, 0, 1, "Print service data", &ngc_print},
    {"quit", 'q', 0, 0, 0, "Stops initng, (HANG SYSTEM)", &ngc_quit},
    {"restart", 'r', 1, 1, 1, "restart service", &ngc_restart_service},
    {"reload", 'R', 1, 0, 1, "reload service data from disk", &ngc_reload},
    {"status", 's', 1, 0, 1, "Output current service status", &ngc_status},
    {"up", 'u', 1, 1, 1, "start service", &ngc_take_up},
    {"verbose", 'v', 0, 0, 1, "Change verbosity level", &ngc_toggle_verbose},
    /*    {"callfd", 'a', 0, 0, 0, "Force initng to close all files", &ngc_c_a_fd}, */
    {"unload-module", 'w', 1, 1, 0, "Unload a module from initng",
     &ngc_m_unload},
    {"initrestart", 'x', 0, 0, 0, "Warm reboot system", &ngc_restart},
    {"stopunneeded", 'y', 0, 0, 0, "Stop all unneeded services", &ngc_s_l},
    {"zap", 'z', 1, 1, 1, "Zap state of service", &ngc_zap},
    {"fast-reaload", 'c', 0, 0, 0, "Fast Reload of initng", &ngc_f_r},
    {NULL, '\0', 0, 0, 0, NULL, NULL}
};

static const char doc[] =
    "\nusage: ngc -[gudfrj] service\n" "       ngc -[lsxqpn]\n\n";


int ngc_hlp(char *opt)
{
    if (opt)
        return (0);
    print_usage(1);
    return (1);
}

/* print usage on screen */
void print_usage(int f)
{
    int i;

    printf("\n\nusage:\n");

    /* print all options, that can have an option */
    printf("%s -[", Argv);
    for (i = 0; o[i].l; i++)
    {
        if (o[i].a && (f || o[i].v))
            putchar(o[i].s);
    }
    printf("] service\n");

    /* print all options, that dont have an option */
    printf("%s -[", Argv);
    for (i = 0; o[i].l; i++)
    {
        if (!o[i].a && (f || o[i].v))
            putchar(o[i].s);
    }
    printf("]\n\n");

    /* print all help */
    for (i = 0; o[i].l; i++)
    {
        if (f || o[i].v)
            printf(" [%c] %s:\t%s\n", o[i].s, o[i].l, o[i].h);
    }

    printf("\n\n");

}

/* open socket to communicate with */
int open_socket()
{
    int len;
    struct sockaddr_un sockname;

    /* Create the socket. */
    sock = socket(PF_UNIX, SOCK_STREAM, 0);
    if (sock < 0)
    {
        printf("Failed to init socket\n");
        return (FALSE);
    }

    /* Bind a name to the socket. */
    sockname.sun_family = AF_UNIX;
    if (debug)
        len = sprintf(sockname.sun_path, SOCKET_FILENAME_TEST);
    else
        len = sprintf(sockname.sun_path, SOCKET_FILENAME_REAL);

    len += sizeof(sockname.sun_family);

    if (connect(sock, (struct sockaddr *) &sockname, len) < 0)
    {
        close(sock);
        printf("Error connecting to socket\n");
        exit(1);
        return (FALSE);
    }
    return (TRUE);
}


/* close pipes */
int close_socket(void)
{
    close(sock);
    sock = -1;
    return (sock);
}

/* send a command */
void send_command(const char *command)
{

    if ((sock < 0) && (open_socket() == FALSE))
        return;

    write(sock, command, strlen(command));

    /* no idea to stress initng */
    usleep(50000);

    return;
}

/* return the fetched string */
char *command_fetch_string(char *got_from_read)
{
    int is_newline = 0;
    char tmp_read;
    char *tmp_ptr;
    int i = 0;
    int max_str_length = 255;

    if (got_from_read)
    {
        free(got_from_read);
        got_from_read = NULL;
    }

    if (sock < 0)
        return NULL;

    while (is_newline == 0)
    {
        if (read(sock, &tmp_read, 1) == 0)
        {
            close_socket();
            break;
        }

        if (got_from_read)
        {
            if (i == max_str_length)
            {
                max_str_length += 255;
                tmp_ptr = calloc(max_str_length, sizeof(char));
                strcpy(tmp_ptr, got_from_read);
                free(got_from_read);
                got_from_read = tmp_ptr;
            }
        }
        else
            got_from_read = calloc(max_str_length, sizeof(char));

        got_from_read[i++] = tmp_read;

        if (tmp_read == '\n')
            is_newline = 1;
    }

    return got_from_read;
}

/* return the fetched return number */
int command_fetch_int(const char *command)
{
    char got_from_read[20];

    send_command(command);

    got_from_read[read(sock, &got_from_read, 19)] = '\0';
    close_socket();
    return (atoi(got_from_read));
}

/* THIS IS MAIN */
int main(int argc, char *argv[])
{
    int cc = 1;
    int i;

    Argv = argv[0];

    /* print banner */
    printf("Next Generation init Control.\n");
    printf("written by Jimmy Wennlund <jimmy.wennlund@gmail.com>\n\n");


    if (argc < 2)
        print_usage(0);

    /* check if cmd-line contains "ngdc", if set debug */
    if (strstr(argv[0], "ngdc"))
    {
        printf("Warning. This is ngdc!\n");
        debug = 1;
    }

    while (cc <= argc && argv[cc])
    {
        /* every fresh start needs a '-' char */
        if (argv[cc][0] != '-')
        {
            bad_syntax();
            exit(1);
        }

        /* handle --option */
        if (argv[cc][1] == '-')
        {                                   /* if its a "--option" */
            int j = 0;
            char *op = NULL;

            /* find the '=' char */
            while (argv[cc][j + 2] && argv[cc][j + 2] != '=')
                j++;

            /* if we found the '=', set op to option */
            if (argv[cc][j + 2] == '=')
                op = &argv[cc][j + 3];


            for (i = 0; o[i].l; i++)
            {
                /* if arg have an option, and use an option */
                if (op &&
                    op[0] != '-' &&
                    op[1] != '-' &&
                    o[i].a &&
                    (int) strlen(o[i].l) == j &&
                    strncasecmp(o[i].l, &argv[cc][2], j) == 0)
                {

                    /* call the funtion listed in struct */
                    (*o[i].exec) (op);

                    cc++;                   /* go to next arg */
                    break;
                }

                /* if arg not have an option and dont need an option */
                if (!op &&
                    !o[i].b &&
                    (int) strlen(o[i].l) == j &&
                    strncasecmp(o[i].l, &argv[cc][2], j) == 0)
                {

                    /* call the funtion listed in struct */
                    (*o[i].exec) (NULL);

                    cc++;                   /* go to next arg */
                    break;
                }
            }


            /* handle -o */
        }
        else
        {

            /* browse list of options */
            for (i = 0; o[i].l; i++)
                /* if this is a match */
                if (argv[cc][1] == o[i].s)
                {
                    if (o[i].a &&           /* can option have an option ? */
                        argv[cc + 1] != NULL && /* is there a string left to be an option ? */
                        argv[cc + 1][0] != '-')
                    {                       /* and is it not an '-' (new option arg) */

                        /* call the funtion listed in struct */
                        (*o[i].exec) (argv[cc + 1]);

                        cc = cc + 2;        /* star over on the one after our argument */
                        break;
                    }
                    else
                    {
                        if (o[i].b)
                        {
                            printf("service \"%s\" needs an option!\n",
                                   o[i].l);
                            bad_syntax();
                            exit(1);
                        }

                        /* call the funtion listed in struct */
                        (*o[i].exec) (NULL);

                        cc++;               /* start over with next argument */
                        break;
                    }
                }

        }
        /* if the end of o struct was reached, then no option was found */
        if (!o[i].l)
        {
            printf("did not find option like %s\n", argv[cc]);
            bad_syntax();
            exit(1);
        }
    }
    exit(0);
}


/* TEMPORARY MOVE THIS!! */
int ngc_c_a_fd(char *service)
{
    char write_this[101];
    int s;

    sprintf(write_this, "a%s", service);
    if ((s = command_fetch_int(write_this)) == 1)
    {
        printf("Returned OK :-)\n");
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed to start process %i :-(\n", s);
        return (FALSE);
    }
}


/* get pid of */
int ngc_get_pid(char *service)
{
    char write_this[101];
    int pid = 0;

    sprintf(write_this, "g%s", service);
    pid = command_fetch_int(write_this);
    if (pid > 0)
    {
        printf("pid of %s is: %i\n", service, pid);
        return (pid);
    }
    else
    {
        printf("ngc: Error getting pid of %s.\n", service);
        return (0);
    }
}

/* start service */
int ngc_take_up(char *service)
{
    char write_this[101];
    int s;

    sprintf(write_this, "u%s", service);
    if ((s = command_fetch_int(write_this)) == 1)
    {
        printf("Process %s started. :-)\n", service);
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed to start process %s. %i :-(\n", service, s);
        return (FALSE);
    }
}

/* take service down */
int ngc_take_down(char *service)
{
    char write_this[101];

    sprintf(write_this, "d%s", service);
    if (command_fetch_int(write_this) == 1)
    {
        printf("Process %s stopped. :-)\n", service);
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed to stop process %s. :-(\n", service);
        return (FALSE);
    }
}

/* restart service */
int ngc_restart_service(char *service)
{
    char write_this[101];

    sprintf(write_this, "r%s", service);
    if (command_fetch_int(write_this) == 1)
    {
        printf("Process %s restarted. :-)\n", service);
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed to restart service %s. :-(\n", service);
        return (FALSE);
    }
}

/* restart from */
int ngc_restart_from(char *service)
{
    char write_this[101];

    sprintf(write_this, "j%s", service);
    if (command_fetch_int(write_this) == 1)
    {
        printf("restart from %s ok. :-)\n", service);
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed to restart from process %s. :-(\n", service);
        return (FALSE);
    }
}

/* zap action */
int ngc_zap(char *service)
{
    char write_this[101];

    sprintf(write_this, "z%s", service);
    if (command_fetch_int(write_this) == 1)
    {
        printf("zap %s ok. :-)\n", service);
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed to zap process %s. :-(\n", service);
        return (FALSE);
    }
}

int ngc_debug_add(char *string)
{
    char write_this[101];

    sprintf(write_this, "i%s", string);
    if (command_fetch_int(write_this) == 1)
    {
        printf("%s added. :-)\n", string);
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed add %s. :-(\n", string);
        return (FALSE);
    }
}

int ngc_debug_del(char *string)
{
    char write_this[101];

    sprintf(write_this, "k%s", string);
    if (command_fetch_int(write_this) == 1)
    {
        printf("%s deleted. :-)\n", string);
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed delete %s. :-(\n", string);
        return (FALSE);
    }
}

int print_status(char *servicename, e_a_status status, struct tm *ts)
{

    /* Print right message by messagetype */
    switch (status)
    {

        case NONE:
            printf("          %s : No status, maybe a new service.\n",
                   servicename);
            break;
        case LOADING:
            printf(" %.2i:%.2i:%.2i %s : loading.\n", ts->tm_hour, ts->tm_min,
                   ts->tm_sec, servicename);
            break;
        case STARTING:
            printf(" %.2i:%.2i:%.2i %s : starting.\n", ts->tm_hour,
                   ts->tm_min, ts->tm_sec, servicename);
            break;
        case WAITING_FOR_START_DEP:
            printf(" %.2i:%.2i:%.2i %s : waiting for dependency to start.\n",
                   ts->tm_hour, ts->tm_min, ts->tm_sec, servicename);
            break;
        case START_DEP_MET:
            printf(" %.2i:%.2i:%.2i %s : start dependency met, starting.\n",
                   ts->tm_hour, ts->tm_min, ts->tm_sec, servicename);
            break;
        case FAIL_STARTING:
            printf(" %.2i:%.2i:%.2i %s : FAILED to start.\n", ts->tm_hour,
                   ts->tm_min, ts->tm_sec, servicename);
            break;
        case START_DEP_FAILED:
            printf
                (" %.2i:%.2i:%.2i %s : a dependency to service FAILED to start.\n",
                 ts->tm_hour, ts->tm_min, ts->tm_sec, servicename);
            break;
        case START_LAUNCHING:
            printf(" %.2i:%.2i:%.2i %s : launching start.\n", ts->tm_hour,
                   ts->tm_min, ts->tm_sec, servicename);
            break;
        case WAITING_FOR_PID_FILE:
            printf(" %.2i:%.2i:%.2i %s : waiting for pid file.\n",
                   ts->tm_hour, ts->tm_min, ts->tm_sec, servicename);
            break;
        case RUNNING:
            printf(" %.2i:%.2i:%.2i %s : running.\n", ts->tm_hour, ts->tm_min,
                   ts->tm_sec, servicename);
            break;
        case DONE:
            printf(" %.2i:%.2i:%.2i %s : done.\n", ts->tm_hour, ts->tm_min,
                   ts->tm_sec, servicename);
            break;
        case RESPAWNING:
            printf(" %.2i:%.2i:%.2i %s : respawning.\n", ts->tm_hour,
                   ts->tm_min, ts->tm_sec, servicename);
            break;
        case STOPPING:
            printf(" %.2i:%.2i:%.2i %s : stopping.\n", ts->tm_hour,
                   ts->tm_min, ts->tm_sec, servicename);
            break;
        case WAITING_FOR_STOP_DEP:
            printf
                (" %.2i:%.2i:%.2i %s : is waiting for dependencies to stop.\n",
                 ts->tm_hour, ts->tm_min, ts->tm_sec, servicename);
            break;
        case STOP_DEP_MET:
            printf
                (" %.2i:%.2i:%.2i %s : all stop dependencies are met, now stopping\n",
                 ts->tm_hour, ts->tm_min, ts->tm_sec, servicename);
            break;
        case STOP_DEP_FAILED:
            printf
                (" %.2i:%.2i:%.2i %s : a dependency to service FAILED to stop.\n",
                 ts->tm_hour, ts->tm_min, ts->tm_sec, servicename);
            break;
        case FAIL_STOPPING:
            printf(" %.2i:%.2i:%.2i %s : stopping.\n", ts->tm_hour,
                   ts->tm_min, ts->tm_sec, servicename);
            break;
        case STOP_LAUNCHING:
            printf
                (" %.2i:%.2i:%.2i %s : stop process is launching.\n",
                 ts->tm_hour, ts->tm_min, ts->tm_sec, servicename);
            break;
        case STOPPED:
            printf(" %.2i:%.2i:%.2i %s : stopped.\n", ts->tm_hour, ts->tm_min,
                   ts->tm_sec, servicename);
            break;
        case STOP_MARKED:
            printf(" %.2i:%.2i:%.2i %s : marked stop.\n", ts->tm_hour,
                   ts->tm_min, ts->tm_sec, servicename);
            break;
        case RESTARTING:
            printf(" %.2i:%.2i:%.2i %s : marked RESTARTING.\n", ts->tm_hour,
                   ts->tm_min, ts->tm_sec, servicename);
            break;
        default:
            printf(" %s : error!\n", servicename);
            return (FALSE);
            break;
    }

    return (TRUE);

}

/* Find out father to */
int ngc_father_to(char *service)
{
    char write_this[101];
    char *got_this = NULL;

    sprintf(write_this, "f%s", service);
    send_command(write_this);

    while ((got_this = command_fetch_string(got_this)))
        printf("%s", got_this);

    return (TRUE);
}

/* List loaded modules */
int ngc_m_list(char *opt)
{
    char *got_this = NULL;

    if (opt)
        return (0);

    send_command("m");
    got_this = command_fetch_string(got_this);

    if (got_this)
        printf("Loaded modules:\n");

    while (got_this)
    {
        printf("%s", got_this);
        got_this = command_fetch_string(got_this);
    }

    return (TRUE);
}

/* load a module */
int ngc_m_load(char *module)
{
    char write_this[101];

    if (!module)
        return (0);
    sprintf(write_this, "o%s", module);
    return (command_fetch_int(write_this));
}

/* unload a module */
int ngc_m_unload(char *module)
{
    char write_this[101];

    if (!module)
        return (0);
    sprintf(write_this, "w%s", module);
    return (command_fetch_int(write_this));
}

/* reload service data from disk */
int ngc_reload(char *service)
{
    char write_this[101];

    sprintf(write_this, "R%s", service);
    if (command_fetch_int(write_this) == 1)
    {
        printf("Service %s data reloaded. :-)\n", service);
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed to reload %s. :-(\n", service);
        return (FALSE);
    }
}

/* get history */
int ngc_history(char *service)
{
    char *got_this = NULL;
    int i = 0;
    int type;
    long int t;
    struct tm *ts;
    char servicename[41];
    char write_this[101];

    /* Fetch input from initng */
    if (service)
    {
        /*printf("\"p%s\"", service); */
        sprintf(write_this, "l%s", service);
        send_command(write_this);
    }
    else
        send_command("l");

    got_this = command_fetch_string(got_this);

    if (!got_this)
        return (FALSE);

    if (strcmp(got_this, "0\n") == 0)
    {
        while ((got_this = command_fetch_string(got_this)))
            printf("%s", got_this);
        return (FALSE);
    }

    /* print head */
    printf(" hh:mm:ss service                                  : status\n");
    printf(" ----------------------------------------------------------\n");

    while (got_this)
    {
        /* Parse the line into type and servicename */
        if (!sscanf(got_this, "%i|%li|%s\n", &type, &t, servicename))
            break;

        time_t test = (time_t) t;

        ts = localtime(&test);

        /*timen=(time_t)t; */

        /* Fill out 40 blanks for nice output */
        for (i = strlen(servicename); i <= 40; i++)
            servicename[i] = ' ';
        servicename[40] = '\0';

        print_status(servicename, type, ts);

        got_this = command_fetch_string(got_this);
    }
    return (TRUE);
}

/* get running processes */
int ngc_status(char *service)
{
    char *got_this = NULL;
    int i = 0;
    long int ti;
    struct tm *ts;
    int type;
    char servicename[41];
    char write_this[101];

    /* Fetch input from initng */
    if (service)
    {
        /*printf("\"p%s\"", service); */
        sprintf(write_this, "s%s", service);
        send_command(write_this);
    }
    else
        send_command("s");

    got_this = command_fetch_string(got_this);
    if (!got_this)
        return (FALSE);

    if (strcmp(got_this, "0\n") == 0)
    {
        while ((got_this = command_fetch_string(got_this)))
            printf("%s", got_this);
        return (FALSE);
    }

    /* print head */
    printf(" hh:mm:ss service                                  : status\n");
    printf(" ----------------------------------------------------------\n");

    while (got_this)
    {
        /* Parse the line into type and servicename */
        if (!sscanf(got_this, "%i|%li|%s\n", &type, &ti, servicename))
            break;

        time_t test = (time_t) ti;

        ts = localtime(&test);

        /* Fill out 40 blanks for nice output */
        for (i = strlen(servicename); i <= 40; i++)
            servicename[i] = ' ';
        servicename[40] = '\0';

        print_status(servicename, type, ts);

        got_this = command_fetch_string(got_this);
    }
    return (TRUE);
}

/* restart initng */
int ngc_restart(char *opt)
{
    if (opt)
        return (0);
    if (command_fetch_int("x") == 1)
    {
        printf("Initng reloading itself now...\n");
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed to reload initng. :-(\n");
        return (FALSE);
    }
}

/* stop unneeded */
int ngc_s_l(char *opt)
{
    if (opt)
        return (0);
    if (command_fetch_int("y") == 1)
    {
        printf("Initng stopping unneeded services...\n");
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed on request. :-(\n");
        return (FALSE);
    }
}

/* restart initng */
int ngc_toggle_verbose(char *opt)
{
    if (opt)
        return (0);
    int a = command_fetch_int("v");

    switch (a)
    {
        case 0:
            printf("Verbose is off, and won't watch g.verbose_this\n");
            break;
        case 1:
            printf("All verbose is on\n");
            break;
        case 2:
            printf("Will verbose whats in g.verbose_this\n");
            break;
        case 3:
            printf("Will verbose all but ! in g.verbose_this\n");
            break;
        default:
            printf("Unknown verbose mode!\n");
            return (FALSE);
            break;
    }
    return (TRUE);
}

/* quit initng (system will hang) */
int ngc_quit(char *opt)
{
    if (opt)
        return (0);
    if (!debug)
    {
        printf("Won't send a quit request to init, it will hang!\n");
        return (1);
    }
    if (command_fetch_int("q") == 1)
    {
        printf("Initng is quitting...\n");
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed to quit initng. :-(\n");
        return (FALSE);
    }
}

/* Fetch Percent */
int ngc_percent(char *opt)
{
    int percent;

    if (opt)
        return (0);
    percent = command_fetch_int("n");
    printf("Percent done %i.\n", percent);
    exit(percent);
}

/* Print services on daemon out */
int ngc_print(char *service)
{
    char write_this[101];
    char *res = NULL;

    if (service)
    {
        /*printf("\"p%s\"", service); */
        sprintf(write_this, "p%s", service);
        send_command(write_this);
    }
    else
        send_command("p");

    while ((res = command_fetch_string(NULL)))
        printf("%s", res);

    return (TRUE);
}

int ngc_print_active(char *service)
{
    char write_this[101];
    char *res = NULL;

    if (service)
    {
        /*printf("\"p%s\"", service); */
        sprintf(write_this, "A%s", service);
        send_command(write_this);
    }
    else
        send_command("A");

    while ((res = command_fetch_string(NULL)))
        printf("%s", res);

    return (TRUE);
}

int ngc_a(char *service)
{
    char write_this[101];
    char *res = NULL;

    if (service)
    {
        /*printf("\"p%s\"", service); */
        sprintf(write_this, "a%s", service);
        send_command(write_this);
    }
    else
        return (FALSE);

    while ((res = command_fetch_string(NULL)))
        printf("%s", res);

    return (TRUE);
}

int ngc_b(char *service)
{
    char write_this[101];
    char *res = NULL;

    if (service)
    {
        /*printf("\"p%s\"", service); */
        sprintf(write_this, "b%s", service);
        send_command(write_this);
    }
    else
        return (FALSE);

    while ((res = command_fetch_string(NULL)))
        printf("%s", res);

    return (TRUE);
}

int ngc_f_r(char *service)
{
    (void) service;
    send_command("c");
    return (0);

}

/* Avoid using goto */
void bad_syntax(void)
{
    printf("Bad syntax!\n\n");
    print_usage(0);
}
