
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "../../src/initng.h"


#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <stdlib.h>
#include <time.h>

#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"

#include "initng_readahead_makecache.h"
#include "initng_readahead_database.h"
#include "initng_readahead_procparser.h"

/*
   Todo: 
   this has to be signal safe, check for EINTR
   cleanup?
   split the files into database and /proc parser stuff
   options to stop logging after gdm is started and create a "base" readahead, so it does readahead on that first, and then the rest AFTER that is started
   dont create cache file on each boot, only on first (this is kinda big :p )


   Take out /var/run, and why the hell did (deleted) end up in there?!?
   /var/run/crond.pid (deleted)

   need to run X seconds after boot is up too, to get more files from gdm/coldplug
   - this needs a new plugin caller in initng, should be fairly easy to do though.
 */

void timer_check_and_scan(void)
{
    D_("Scan proc pulse (%d)\n", time(NULL));
    scan_proc();
}

void timer_shutdown(void)
{
    D_("Shutdown timer at %d\n", time(NULL));
    module_unload();
}

void is_system_up(h_sys_state state)
{
    if (state == STATE_UP)
    {
        D_("Shutting down readahead timer in %d seconds\n",
           READAHEAD_TIMEOUT);
        initng_add_hook(TIMERS, 90, &timer_shutdown, READAHEAD_TIMEOUT);
    }
    if (state == STATE_STOPPING)
        module_unload();
}

int module_init(const char *version)
{
    if (!g.i_am_init)
    {
        W_("No i_am_init debugging for initng_readahead (yet)\n");
    }
    proc_db = NULL;
    D_("module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module, is compiled for \"%s\" version, and initng is compiled on \"%s\" version, wont load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    W_("Making readahead cache for next boot, this boot will be slower than normal!\n");
    initng_add_hook(SWATCHERS, 90, &is_system_up);
    initng_add_hook(TIMERS, 90, &timer_check_and_scan, 5);
    return (TRUE);
}

void module_unload(void)
{
    D_("module_unload();\n");
    initng_del_hook(SWATCHERS, &is_system_up);
    initng_del_hook(TIMERS, &timer_check_and_scan);
    initng_del_hook(TIMERS, &timer_shutdown);
    if (!proc_db)
        return;

    D_("Saving readahead cache to file\n");
    D_("Cleaning database, resolving links and removing duplicates\n");
    clean_database();
    D_("Loading old database and merging the data\n");
    load_database();
    D_("Checking if database should be trimmed\n");
    boots_since_last_trim++;
    if (boots_since_last_trim >= TRIM_EVERY_X_BOOT)
    {
        D_("TRIMMING DATABASE\n");
        trim_database();
        boots_since_last_trim = 0;
    }
    D_("Sorting database\n");
    sort_database();
    W_("Saving new database\n");
    save_database();

    free_database();

}
